<?php
namespace Teknisa\CADEST\Modules\Produto\Services;

use Teknisa\CADEST\Modules\Produto\Factories\Produto as ProdutoFactory;
use Teknisa\CADEST\Modules\Produto\Services\ProdutoUtilities;
use Teknisa\Libs\Util\Environment;
use Zeedhi\Framework\DB\StoredProcedure\Param;
use Zeedhi\Framework\DB\StoredProcedure\StoredProcedure;
use Doctrine\ORM\EntityManager;

class Produto {

    protected $produtoUtilities;
    protected $produtoFactory;
    
    const PROCEDURE_INSERE_AGRUPA = 'INSERE_AGRUPA';

    public function __construct(ProdutoFactory $produtoFactory, ProdutoUtilities $produtoUtilities, Environment $environment, $nameProvider, EntityManager $entityManager) {
        $this->produtoUtilities = $produtoUtilities;
        $this->produtoFactory   = $produtoFactory;
        $this->environment      = $environment;
        $this->nameProvider     = $nameProvider;
        $this->entityManager    = $entityManager;
    }

    public function alreadyRelatedToTpTree($cdproduto, $nrtparvprod) {
        $params = array('CDPRODUTO' => $cdproduto, 
                        'NRORG' => $this->produtoUtilities->getNrorg());
        $types = $this->produtoFactory->getTpTreeProduct($params);
        foreach ($types as $type) {
            if ($type['NRTPARVPROD'] == $nrtparvprod) {
                throw new \Exception('O produto já se encontra cadastrado para outro grupo no mesmo tipo de árvore.', 1);
            }
        }
    }

    public function applyRulesToCompraType($row) {
        if (!isset($row['CDPRODESTO'])) {
            throw new \Exception('Erro. Produto de Estoque é obrigatório para um produto de compra.',1);
        } else if ($row['CDPRODUTO'] === $row['CDPRODESTO']) {
            throw new \Exception('Erro. Produto de compra com código do produto iqual ao produto de estoque.',1);
        }

        if (!isset($row['VRFATOCONV'])) {
            throw new \Exception('Erro. Fator de Conversão é obrigatório para um produto de compra.',1);
        }else if($row['VRFATOCONV'] <= 0){
            throw new \Exception('Erro. Fator de Conversão deve ser maior que 0.',1);
        }
        
        $row['VRPESOUNID'] = null;
        $row['VRPEFATOCORR'] = null;
        $row['CDPRODPROD'] = null;
        $row['VRFATOCONVPR'] = null;
        $row['IDPRODPROD'] = null;
        return $row;
    }
    
    public function existBarcode($cdproduto, $cdbarproduto) {
        $params = array('CDBARPRODUTO'=> $cdbarproduto,
                        'CDPRODUTO' => $cdproduto,
                        'NRORG'=> $this->produtoUtilities->getNrorg());
                        
        $othersBarcode = $this->produtoFactory->getProductCdbarproduto($params);
        
        if ($othersBarcode !== false) {
            
            throw new \Exception('Erro. Código de Barras já cadastrado para o produto '.$othersBarcode['CDARVPROD'],1);
        }
    }

    public function applyRulesToEstoqueType($row) {
        if (isset($row['CDPRODPROD']) && !isset($row['VRFATOCONVPR'])) {
            throw new \Exception('Erro. É necessário informar o Fator de Conversão Produção quando se seleciona um Produto de Produção.',1);
        }

        $row['VRFATOCONV'] = 1;
        $row['CDPRODESTO'] = $row['CDPRODUTO'];
        return $row;
    }
 
    public function arvprodContainsProdnivante($cdarvprod, $cdarvprodnivante) {
        if (!$this->isLevelOne($cdarvprod)) {
            $params = array('CDARVPROD'=>$cdarvprodnivante,
                            'NRORG'=>$this->produtoUtilities->getNrorg());
            $cdarvprodnivante = $this->produtoFactory->getArvprodByCdarvprod($params);
            $newCdarvprod = str_replace(".", "", $cdarvprod);
            if (!isset($cdarvprodnivante)) {
                throw new \Exception('Erro. Não existe o produto pai informado.', 1);
            } else if (substr($newCdarvprod, 0, (strlen($cdarvprodnivante))) !== $cdarvprodnivante) {
                throw new \Exception('Erro. O código de Produto não está de acordo com o produto pai selecionado.', 1);
            }
        }
    }

    public function canBeDeletedTree($cdarvprod, $cdproduto) {
        $params = array('CDARVPROD'=>$cdarvprod,
                        'NRORG'=>$this->produtoUtilities->getNrorg());
        $children = $this->produtoFactory->getChildren($params);
        
        if (count($children) > 0) {
            throw new \Exception('Exclusão não permitida. Existem produtos relacionados.', 1);
        }
        
    }

    public function correctLevelByArvprod($cdarvprod, $cdnvproduto) {
        $levelByCdarvprod = $this->getCdnvprodutoByCdarvprod($cdarvprod);
        if ($levelByCdarvprod != $cdnvproduto) {
            throw new \Exception('Erro. O nível do produto não corresponde ao Código de Árvore.', 1);
        }
    }

    public function existArvprod($cdarvprod) {
        $params = array('CDARVPROD'=>$cdarvprod,
                        'NRORG'=>$this->produtoUtilities->getNrorg());
        $othersProducts = $this->produtoFactory->getProductByArvprod($params);
        
        if ($othersProducts !== false) {
            throw new \Exception('Erro. Código de Produto já cadastrado para outro produto.',1);
        }
    }

    public function existProduct($cdarvprod) {
        $params = array('CDPRODUTO'=>$cdarvprod,
                        'NRORG'=>$this->produtoUtilities->getNrorg());
        $product = $this->produtoFactory->getProductByCode($params);
        
        if ($product) {
            return true;
        } else {
            return false;
        }
    }

    public function getProductMask() {
        $params = array('NRORG'=>$this->produtoUtilities->getNrorg());
        return $this->produtoFactory->getProductMask($params);
    }

    public function getCdnvprodutoByCdarvprod($cdarvprod) {
        $nv = 1;
        $mask = $this->getProductMask();
        $mask = explode('.', $mask);
        $cdarvprod = str_replace('.', '', $cdarvprod);
        
        for ($i=0; $i < count($mask); $i++) { 
            $sessionLenght = strlen($mask[$i]);
            $arvprodLenght = strlen($cdarvprod);
            if ($sessionLenght === $arvprodLenght) {
                $nv = $nv + $i;
                $i = count($mask);
            }
            $cdarvprod = substr($cdarvprod, ($sessionLenght - $arvprodLenght));
        }
        return $nv;
    }

    public function getNewCdproduto($cdarvprod) {
        $gereratedCode =  $this->produtoUtilities->newCode();
        
        if (strlen($cdarvprod) <= strlen($gereratedCode) && !$this->existProduct($cdarvprod)) {
            return $cdarvprod;
        } else {
            return $gereratedCode;
        }
    }

    public function getSugestedCdarvprod($parentcdproduto, $parentnvproduto) {
        $params = array('CDARVPROD'=>$parentcdproduto,
                        'NRORG'=>$this->produtoUtilities->getNrorg());
        $children = $this->produtoFactory->getChildren($params);
        $mask = $this->getProductMask();
        $mask = explode('.', $mask);
        $sessionMask = $mask[intval($parentnvproduto)];
        $zeroFill = strlen($sessionMask);
        if (count($children) === 0) {
            $cdarvprod = str_pad('0', $zeroFill, "0", STR_PAD_LEFT);
        } else {
            $brothersArvprod = array_map(function($brother) {
                return $brother['CDARVPROD'];
            }, $children);
            for ($i=0; $i < count($children); $i++) { 
                $sessionBrotherCdarvprod = substr($children[$i]['CDARVPROD'], (strlen($sessionMask) * (-1)));
                $cdarvprod = $this->incrementArvProd($sessionBrotherCdarvprod, $sessionMask);
                if ($this->respectMask($sessionMask, $cdarvprod) && !in_array($cdarvprod, $brothersArvprod)) {
                    $i = count($children);
                }
            }
        }
        
        $parentarvprod = $this->produtoFactory->getArvprodByCdarvprod($params);
        $cdarvprod = $parentarvprod.$cdarvprod;
                        
        return $cdarvprod;
    }

    private function isLevelOne($cdarvprod) {
        $level = $this->getCdnvprodutoByCdarvprod($cdarvprod);
        if ($level === 1) {
            return true;
        } else {
            return false;
        }
    }
    
    private function imcrementOne($val, $mask) {
        if ($mask === 'a') {
            if ($val == 9) {
                return 'A';
            } else if ($val === 'Z') {
                return '0';
            } else {
                return ++$val;
            }
        } else {
            $val++;
            if (strlen($val) != strlen($mask)) {
                return '0';
            } else {
                return $val;
            }
        }
    }
    
    private function incrementArvProd($arvprod, $mask) {
        $lastCaracterPostion = strlen($arvprod)- 1;
        $explodedArvprod = str_split($arvprod);
        $explodedMask = str_split($mask);
        
        $explodedArvprod[$lastCaracterPostion] = $this->imcrementOne($explodedArvprod[$lastCaracterPostion], $explodedMask[$lastCaracterPostion]);
        $lastCarcter = $explodedArvprod[$lastCaracterPostion];

        if ($explodedArvprod[$lastCaracterPostion] == 0 && strlen($arvprod) > 1) {
            array_pop($explodedArvprod);
            array_pop($explodedMask);
            $arvprod = implode("", $explodedArvprod);
            $mask = implode("", $explodedMask);
            $arvprod = $this->incrementArvProd($arvprod, $mask).$lastCarcter;
        } else {
            $arvprod = implode("", $explodedArvprod);
        }
        
        return $arvprod;
    }

    public function newProductType($product) {
        if (isset($product['IDESTOCOMP'])) {
            return $product['IDESTOCOMP'];
        } else {
            throw new \Exception('Erro. Não foi informado se o produto é de estoque ou de compra.',1);
        }
    }

    public function productType($product) {
        $params = array(
            'CDPRODUTO'=>$product['CDPRODUTO'],
            'NRORG'=>$this->produtoUtilities->getNrorg()
        );
        $productInDbType = $this->produtoFactory->getProductType($params);
        return $productInDbType;
    }

    public function removeMask($cdarvprodWithMask) {
        $cdarvprod = '';
        $cdarvprodWithMask = explode('.', $cdarvprodWithMask);
        foreach ($cdarvprodWithMask as $value) {
            $cdarvprod = $cdarvprod.$value;
        }

        return strtoupper($cdarvprod);
    }

    private function respectMask($mask, $value) {
        if (strlen($mask) === strlen($value)) {
            return true;
        } else {
            return false;
        }
    }

    public function validateType($cdproduto, $supposedType) {
        $params = array(
            'CDPRODUTO'=>$cdproduto,
            'NRORG'=>$this->produtoUtilities->getNrorg()
        );
        $productInDbType = $this->produtoFactory->getProductType($params);
        if ($productInDbType !== $supposedType) {
            throw new \Exception('Erro. Houve alteração no tipo do protudo, essa edição não é permitida.',1);
        }
    }
    
    public function addNewParamimpprod() {
        $this->produtoFactory->addNewParamimpprod();
    }
    
    public function getDataSetProdutoFilial($cdproduto, $cdOperador, $nrorg) {
        $filiais = $this->produtoFactory->getFilial($cdproduto);

        $prodfili = array();
        
        $count = 0;
        
        $dataSource = $this->nameProvider->getDataSourceByName("prodfili_insert");
        
        $produtoBase = array("CDPRODUTO" => $cdproduto, "NRORG" => $nrorg);
        $this->adicionarCamposAuditoria($produtoBase, $cdOperador, $nrorg);
        $produtoBase = $this->getRowWithDefaultColumnsProperties($produtoBase, $dataSource);
        
        $produtoBase["CDALMOXPRODPL"] = " ";
        $produtoBase["CDLOCPRODPL"]   = " ";
        $produtoBase["CDALMOXPROD"]   = " ";
        $produtoBase["CDLOCPROD"]     = " ";
        $produtoBase["CDALMOXREPROV"] = " ";
        $produtoBase["CDLOCALREPROV"] = " ";
        
        foreach ($filiais as $filial) {
            $produto = $produtoBase;
            $produto["CDFILIAL"] = $filial["CDFILIAL"];
            
            if(isset($prodfili[$count])) {
                if(count($prodfili[$count]) > 250) {
                    $count++;
                    $prodfili[$count] = array();
                    array_push($prodfili[$count], $produto);
                }
                else {
                    array_push($prodfili[$count], $produto);
                }
            }
            else {
                $prodfili[$count] = array($produto);
            }
            count($prodfili[$count]);
        }
        return $prodfili;
    }
    
    protected function adicionarCamposAuditoria(&$row, $cdOperador, $nrOrg) {
        $currentTime = $this->environment->getCurrentTime()->format("d/m/Y H:i:s");
        $row['NRORG'] = $nrOrg;
        $row['IDATIVO'] = isset($row['IDATIVO']) ? ($row['IDATIVO'] ?: 'S') : 'S';
        $row['DTULTATU'] = $currentTime;
        $row["__is_new"] = true;
        $row['DTINCLUSAO'] = $currentTime;
        $row['NRORGINCLUSAO'] = $nrOrg;
        $row['CDOPERINCLUSAO'] = $cdOperador;
        $row['NRORGULTATU']  = $nrOrg;
        $row['CDOPERULTATU'] = $cdOperador;
        $row['DTULTATU']     = $currentTime;
    }

    private function getRowWithDefaultColumnsProperties($row, $dataSource) {

        if (method_exists($dataSource, 'getColumnsProperties')) {
            $columnsProperties = $dataSource->getColumnsProperties();

            if (isset($columnsProperties)) {
                foreach ($columnsProperties as $columnProperty) {
                    if (!array_key_exists($columnProperty['COLUMN_NAME'], $row)) {
                        if ($columnProperty['NULLABLE'] == 'Y') {
                            $row[$columnProperty['COLUMN_NAME']] = $this->getDataDefaultValue($columnProperty, $columnProperty["COLUMN_NAME"]);
                        } else if ($columnProperty['COLUMN_NAME'] == 'CDFILIAL') {
                            $row[$columnProperty['COLUMN_NAME']] = $this->environment->getCdFilial();
                        } else if ($columnProperty['NULLABLE'] == 'N') {
                            if (isset($columnProperty['DATA_DEFAULT'])){
                                $row[$columnProperty['COLUMN_NAME']] = $this->getDataDefaultValue($columnProperty, $columnProperty["COLUMN_NAME"]);
                            } else {
                                throw new \Exception("Column ".$columnProperty['COLUMN_NAME']." não pode ser nula e não possui um valor default ! Insira um valor para este campo.", 1);
                            }
                        }
                    }
                }
            }
        }

        return $row;
    }
    
    
    public function deleteProduct($row){
        
        try{
            $connection = $this->entityManager->getConnection();
            $connection->beginTransaction();
            $cdproduto = $row['CDPRODUTO'];
            $nrorg = $this->produtoUtilities->getNrorg();
            $this->produtoFactory->deleteProdfili($cdproduto, $nrorg);
            $this->produtoFactory->deleteProdfabric($cdproduto, $nrorg);
            $this->produtoFactory->deleteProdprinativ($cdproduto, $nrorg);
            $this->produtoFactory->deleteProdforncomp($cdproduto, $nrorg);
            $livrosMag = $this->produtoFactory->getLivroMagnetico();
            
            foreach($livrosMag as $livro){
                $this->produtoFactory->deleteProdRelacproditgen($livro['CDLIVROMAG'], $cdproduto, $nrorg);
            }
            
            $this->produtoFactory->deleteProduto($cdproduto, $nrorg);
            
            $connection->commit();
        }catch(\Exception $e){
            $connection->rollBack();
            if(strpos($e->getMessage(), 'ORA-02292') === false){
				throw  new \Exception($e->getMessage());
			} else {
				throw new \Exception("Exclusão bloqueada! Existem registros filhos na base de dados que impedem a exclusão do produto.");
			}
        }
        
    }

    private function getDataDefaultValue($columnProperty, $columnName) {
        if(strpos($columnName, "CDALMOX") !== false || strpos($columnName, "CDLOC") !== false) {
            return " ";
        }
        if ($columnProperty['DATA_DEFAULT'] == null) {
            return null;
        } else {
            switch ($columnProperty['DATA_TYPE']) {
                case 'DATE':
                    return $this->environment->getCurrentTime()->format("d/m/Y H:i:s");
                    break;
                case 'NUMBER':
                    return intval($columnProperty['DATA_DEFAULT']);
                    break;
                case 'FLOAT':
                    return floatval($columnProperty['DATA_DEFAULT']);
                    break;
                default:
                    return $columnProperty['DATA_DEFAULT'];
                    break;
            }
        }
    }
    
     public function callInsereAgrupa($cdarvprod, $stEdit = 'N') {
        $procedure = new StoredProcedure($this->entityManager->getConnection(), self::PROCEDURE_INSERE_AGRUPA);
        
        $procedure->addParam(new Param('P_CDARVPROD', Param::PARAM_INPUT, $cdarvprod,       Param::PARAM_TYPE_STR));
        $procedure->addParam(new Param('P_stEdit',    Param::PARAM_INPUT, $stEdit,          Param::PARAM_TYPE_STR));
   
        $procedure->execute();
    }
}